@@ -1,12 +1,14 @@
 // Program.cpp (Oclgrind)
-// Copyright (c) 2013-2015, James Price and Simon McIntosh-Smith,
+// Copyright (c) 2013-2016, James Price and Simon McIntosh-Smith,
 // University of Bristol. All rights reserved.
 //
 // This program is provided under a three-clause BSD license. For full
 // license terms please see the LICENSE file distributed with this
 // source code.
 
+#include "config.h"
 #include "common.h"
+
 #include <fstream>
 
 #if defined(_WIN32) && !defined(__MINGW32__)
@@ -29,7 +31,9 @@
 #include "clang/CodeGen/CodeGenAction.h"
 #include "clang/Frontend/CompilerInstance.h"
 #include "clang/Frontend/TextDiagnosticPrinter.h"
+#include "clang/Lex/PreprocessorOptions.h"
 
+#include "Context.h"
 #include "Kernel.h"
 #include "Program.h"
 #include "WorkItem.h"
@@ -129,7 +133,11 @@ bool Program::build(const char *options, list<Header> headers)
   args.push_back("-cl-std=CL1.2");
   args.push_back("-cl-kernel-arg-info");
   args.push_back("-fno-builtin");
+#if LLVM_VERSION >= 38
+  args.push_back("-debug-info-kind=standalone");
+#else
   args.push_back("-g");
+#endif
   args.push_back("-triple");
   if (sizeof(size_t) == 4)
     args.push_back("spir-unknown-unknown");
@@ -150,6 +158,10 @@ bool Program::build(const char *options, list<Header> headers)
   bool optimize = true;
   bool cl12     = true;
 
+  // Disable optimizations by default if in interactive mode
+  if (checkEnv("OCLGRIND_INTERACTIVE"))
+    optimize = false;
+
   // Add OpenCL build options
   const char *mainOptions = options;
   const char *extraOptions = getenv("OCLGRIND_BUILD_OPTIONS");
@@ -164,7 +176,8 @@ bool Program::build(const char *options, list<Header> headers)
     // Ignore options that break PCH
     if (strcmp(opt, "-cl-fast-relaxed-math") != 0 &&
         strcmp(opt, "-cl-finite-math-only") != 0 &&
-        strcmp(opt, "-cl-single-precision-constant") != 0)
+        strcmp(opt, "-cl-single-precision-constant") &&
+        strcmp(opt, "-cl-unsafe-math-optimizations") != 0)
     {
       // Check for optimization flags
       if (strcmp(opt, "-O0") == 0 || strcmp(opt, "-cl-opt-disable") == 0)
@@ -178,6 +191,12 @@ bool Program::build(const char *options, list<Header> headers)
         continue;
       }
 
+#if LLVM_VERSION >= 37
+      // Clang no longer supports -cl-no-signed-zeros
+      if (strcmp(opt, "-cl-no-signed-zeros") == 0)
+        continue;
+#endif
+
       // Check for -cl-std flag
       if (strncmp(opt, "-cl-std=", 8) == 0)
       {
@@ -214,11 +233,11 @@ bool Program::build(const char *options, list<Header> headers)
 #if defined(_WIN32) && !defined(__MINGW32__)
       char libpath[4096];
       HMODULE dll;
-      if (GetModuleHandleEx(
+      if (GetModuleHandleExA(
             GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS |
             GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
             (LPCSTR)&Program::createFromBitcode, &dll) &&
-          GetModuleFileName(dll, libpath, sizeof(libpath)))
+          GetModuleFileNameA(dll, libpath, sizeof(libpath)))
       {
 #else
       Dl_info dlinfo;
@@ -274,6 +293,7 @@ bool Program::build(const char *options, list<Header> headers)
 
     args.push_back("-include-pch");
     args.push_back(pch);
+    args.push_back("-fno-validate-pch");
   }
   else
   {
@@ -325,8 +345,7 @@ bool Program::build(const char *options, list<Header> headers)
   compiler.getPreprocessorOpts().addRemappedFile(REMAP_INPUT, buffer.release());
 
   // Compile
-  llvm::LLVMContext& context = llvm::getGlobalContext();
-  clang::EmitLLVMOnlyAction action(&context);
+  clang::EmitLLVMOnlyAction action(m_context->getLLVMContext());
   if (compiler.ExecuteAction(action))
   {
     // Retrieve module
@@ -338,32 +357,34 @@ bool Program::build(const char *options, list<Header> headers)
       stripDebugIntrinsics();
     }
 
-    // Initialize pass managers
-    llvm::legacy::PassManager modulePasses;
-    llvm::legacy::FunctionPassManager functionPasses(m_module.get());
-#if LLVM_VERSION < 37
-    modulePasses.add(new llvm::DataLayoutPass());
-    functionPasses.add(new llvm::DataLayoutPass());
-#endif
-
     // Run optimizations on module
     if (optimize)
     {
+      // Initialize pass managers
+      llvm::legacy::PassManager modulePasses;
+      llvm::legacy::FunctionPassManager functionPasses(m_module.get());
+#if LLVM_VERSION < 37
+      modulePasses.add(new llvm::DataLayoutPass());
+      functionPasses.add(new llvm::DataLayoutPass());
+#endif
+
       // Populate pass managers with -Oz
       llvm::PassManagerBuilder builder;
       builder.OptLevel = 2;
       builder.SizeLevel = 2;
       builder.populateModulePassManager(modulePasses);
       builder.populateFunctionPassManager(functionPasses);
+
+      // Run passes
+      functionPasses.doInitialization();
+      llvm::Module::iterator fItr;
+      for (fItr = m_module->begin(); fItr != m_module->end(); fItr++)
+        functionPasses.run(*fItr);
+      functionPasses.doFinalization();
+      modulePasses.run(*m_module);
     }
 
-    // Run passes
-    functionPasses.doInitialization();
-    llvm::Module::iterator fItr;
-    for (fItr = m_module->begin(); fItr != m_module->end(); fItr++)
-      functionPasses.run(*fItr);
-    functionPasses.doFinalization();
-    modulePasses.run(*m_module);
+    removeLValueLoads();
 
     m_buildStatus = CL_BUILD_SUCCESS;
   }
@@ -448,14 +469,22 @@ Program* Program::createFromBitcode(const Context *context,
   }
 
   // Parse bitcode into IR module
+#if LLVM_VERSION < 37
   llvm::ErrorOr<llvm::Module*> module =
-    parseBitcodeFile(buffer->getMemBufferRef(), llvm::getGlobalContext());
+#else
+  llvm::ErrorOr<unique_ptr<llvm::Module>> module =
+#endif
+    parseBitcodeFile(buffer->getMemBufferRef(), *context->getLLVMContext());
   if (!module)
   {
     return NULL;
   }
 
+#if LLVM_VERSION < 37
   return new Program(context, module.get());
+#else
+  return new Program(context, module.get().release());
+#endif
 }
 
 Program* Program::createFromBitcodeFile(const Context *context,
@@ -470,35 +499,52 @@ Program* Program::createFromBitcodeFile(const Context *context,
   }
 
   // Parse bitcode into IR module
+#if LLVM_VERSION < 37
   llvm::ErrorOr<llvm::Module*> module =
+#else
+  llvm::ErrorOr<unique_ptr<llvm::Module>> module =
+#endif
     parseBitcodeFile(buffer->get()->getMemBufferRef(),
-                     llvm::getGlobalContext());
+                     *context->getLLVMContext());
   if (!module)
   {
     return NULL;
   }
 
+#if LLVM_VERSION < 37
   return new Program(context, module.get());
+#else
+  return new Program(context, module.get().release());
+#endif
 }
 
 Program* Program::createFromPrograms(const Context *context,
                                      list<const Program*> programs)
 {
   llvm::Module *module = new llvm::Module("oclgrind_linked",
-                                          llvm::getGlobalContext());
+                                          *context->getLLVMContext());
+#if LLVM_VERSION < 38
   llvm::Linker linker(module);
+#else
+  llvm::Linker linker(*module);
+#endif
 
   // Link modules
   list<const Program*>::iterator itr;
   for (itr = programs.begin(); itr != programs.end(); itr++)
   {
-    if (linker.linkInModule(CloneModule((*itr)->m_module.get())))
+#if LLVM_VERSION < 38
+    llvm::Module *m = llvm::CloneModule((*itr)->m_module.get());
+#else
+    unique_ptr<llvm::Module> m = llvm::CloneModule((*itr)->m_module.get());
+#endif
+    if (linker.linkInModule(std::move(m)))
     {
       return NULL;
     }
   }
 
-  return new Program(context, linker.getModule());
+  return new Program(context, module);
 }
 
 Kernel* Program::createKernel(const string name)
@@ -508,7 +554,7 @@ Kernel* Program::createKernel(const string name)
 
   // Iterate over functions in module to find kernel
   llvm::Function *function = NULL;
-
+#if LLVM_VERSION < 37
   // Query the SPIR kernel list
   llvm::NamedMDNode* tuple = m_module->getNamedMetadata("opencl.kernels");
   // No kernels in module
@@ -538,6 +584,17 @@ Kernel* Program::createKernel(const string name)
       break;
     }
   }
+#else
+  for (auto F = m_module->begin(); F != m_module->end(); F++)
+  {
+    if (F->getCallingConv() == llvm::CallingConv::SPIR_KERNEL &&
+        F->getName() == name)
+    {
+      function = &*F;
+      break;
+    }
+  }
+#endif
 
   if (function == NULL)
   {
@@ -566,20 +623,17 @@ Kernel* Program::createKernel(const string name)
   }
 }
 
-unsigned char* Program::getBinary() const
+void Program::getBinary(unsigned char *binary) const
 {
   if (!m_module)
-  {
-    return NULL;
-  }
+    return;
 
   std::string str;
   llvm::raw_string_ostream stream(str);
   llvm::WriteBitcodeToFile(m_module.get(), stream);
   stream.str();
-  unsigned char *bitcode = new unsigned char[str.length()];
-  memcpy(bitcode, str.c_str(), str.length());
-  return bitcode;
+
+  memcpy(binary, str.c_str(), str.length());
 }
 
 size_t Program::getBinarySize() const
@@ -632,6 +686,7 @@ list<string> Program::getKernelNames() const
 {
   list<string> names;
 
+#if LLVM_VERSION < 37
   // Query the SPIR kernel list
   llvm::NamedMDNode* tuple = m_module->getNamedMetadata("opencl.kernels");
 
@@ -656,6 +711,15 @@ list<string> Program::getKernelNames() const
       names.push_back(kernelFunction->getName());
     }
   }
+#else
+  for (auto F = m_module->begin(); F != m_module->end(); F++)
+  {
+    if (F->getCallingConv() == llvm::CallingConv::SPIR_KERNEL)
+    {
+      names.push_back(F->getName());
+    }
+  }
+#endif
 
   return names;
 }
@@ -664,6 +728,7 @@ unsigned int Program::getNumKernels() const
 {
   assert(m_module);
 
+#if LLVM_VERSION < 37
   // Extract kernels from metadata
   llvm::NamedMDNode* tuple = m_module->getNamedMetadata("opencl.kernels");
 
@@ -672,6 +737,19 @@ unsigned int Program::getNumKernels() const
     return 0;
 
   return tuple->getNumOperands();
+#else
+  unsigned int num = 0;
+
+  for (auto F = m_module->begin(); F != m_module->end(); F++)
+  {
+    if (F->getCallingConv() == llvm::CallingConv::SPIR_KERNEL)
+    {
+      num++;
+    }
+  }
+
+  return num;
+#endif
 }
 
 const string& Program::getSource() const
@@ -697,13 +775,303 @@ unsigned long Program::getUID() const
   return m_uid;
 }
 
+void Program::pruneDeadCode(llvm::Instruction *instruction)
+{
+  // Remove instructions that have no uses
+  if (instruction->getNumUses() == 0)
+  {
+    // Get list of operands
+    set<llvm::Value*> operands;
+    {
+      llvm::Instruction::op_iterator op;
+      for (op = instruction->op_begin(); op != instruction->op_end(); op++)
+      {
+        operands.insert(*op);
+      }
+    }
+
+    // Remove instruction
+    instruction->eraseFromParent();
+
+    // Prune operands
+    set<llvm::Value*>::iterator op;
+    for (op = operands.begin(); op != operands.end(); op++)
+    {
+      if (auto inst = llvm::dyn_cast<llvm::Instruction>(*op))
+        pruneDeadCode(inst);
+    }
+  }
+}
+
+void Program::removeLValueLoads()
+{
+  // Get list of aggregate store instructions
+  set<llvm::StoreInst*> aggStores;
+  for (llvm::Module::iterator F = m_module->begin(); F != m_module->end(); F++)
+  {
+    llvm::Function *f = &*F;
+    for (llvm::inst_iterator I = inst_begin(f), E = inst_end(f); I != E; I++)
+    {
+      if (auto store = llvm::dyn_cast<llvm::StoreInst>(&*I))
+        aggStores.insert(store);
+    }
+  }
+
+  // Replace aggregate modify-write sequences with direct scalar writes
+  set<llvm::StoreInst*>::iterator itr;
+  for (itr = aggStores.begin(); itr != aggStores.end(); itr++)
+  {
+    scalarizeAggregateStore(*itr);
+  }
+}
+
+void Program::scalarizeAggregateStore(llvm::StoreInst *store)
+{
+  llvm::IntegerType *gepIndexType = (sizeof(size_t)==8) ?
+      llvm::Type::getInt64Ty(m_module.get()->getContext()) :
+      llvm::Type::getInt32Ty(m_module.get()->getContext());
+
+  llvm::Value *storeValue = store->getValueOperand();
+  llvm::Value *vectorPtr  = store->getPointerOperand();
+
+  if (auto insert = llvm::dyn_cast<llvm::InsertElementInst>(storeValue))
+  {
+    llvm::Value *vector = insert->getOperand(0);
+    llvm::Value *value  = insert->getOperand(1);
+    llvm::Value *index  = insert->getOperand(2);
+
+    // Create GEP for scalar value
+    llvm::GetElementPtrInst *scalarPtr = NULL;
+    if (auto gep = llvm::dyn_cast<llvm::GetElementPtrInst>(vectorPtr))
+    {
+      // Create GEP from existing GEP
+      std::vector<llvm::Value*> indices;
+      for (auto idx = gep->idx_begin(); idx != gep->idx_end(); idx++)
+      {
+        indices.push_back(*idx);
+      }
+      indices.push_back(index);
+      scalarPtr = llvm::GetElementPtrInst::Create(
+#if LLVM_VERSION > 36
+        gep->getPointerOperandType()->getPointerElementType(),
+#endif
+        gep->getPointerOperand(), indices);
+    }
+    else
+    {
+      // Create GEP from non-GEP pointer
+      std::vector<llvm::Value*> indices;
+      indices.push_back(llvm::ConstantInt::getSigned(gepIndexType, 0));
+      indices.push_back(index);
+      scalarPtr = llvm::GetElementPtrInst::Create(
+#if LLVM_VERSION > 36
+        vectorPtr->getType()->getPointerElementType(),
+#endif
+        vectorPtr, indices);
+    }
+    scalarPtr->setDebugLoc(store->getDebugLoc());
+    scalarPtr->insertAfter(store);
+
+    // Create direct scalar store
+    llvm::StoreInst *scalarStore = new llvm::StoreInst(
+      value, scalarPtr, store->isVolatile(),
+      getTypeAlignment(value->getType()));
+    scalarStore->setDebugLoc(store->getDebugLoc());
+    scalarStore->insertAfter(scalarPtr);
+
+    // Check if the input to the insertelement instruction came from something
+    // other than a load to the same address as the store
+    llvm::LoadInst *load = llvm::dyn_cast<llvm::LoadInst>(vector);
+    if (!(load && load->getPointerOperand() == store->getPointerOperand()))
+    {
+      // Replace value in store with the input to the insertelement instruction
+      llvm::StoreInst *_store = new llvm::StoreInst(
+        vector, store->getPointerOperand(),
+        store->isVolatile(), store->getAlignment());
+      _store->setDebugLoc(store->getDebugLoc());
+      _store->insertAfter(store);
+
+      // Repeat process with new store
+      if (_store)
+        scalarizeAggregateStore(_store);
+    }
+
+    // Remove vector store and any dead code
+    store->eraseFromParent();
+    pruneDeadCode(insert);
+  }
+  else if (auto shuffle = llvm::dyn_cast<llvm::ShuffleVectorInst>(storeValue))
+  {
+    llvm::Value *v1      = shuffle->getOperand(0);
+    llvm::Value *v2      = shuffle->getOperand(1);
+    llvm::Constant *mask = shuffle->getMask();
+    unsigned maskSize    = mask->getType()->getVectorNumElements();
+
+    // Check if shuffle sources came from a load with same address as the store
+    llvm::LoadInst *load;
+    bool v1SourceIsDest = false, v2SourceIsDest = false;
+    if ((load = llvm::dyn_cast<llvm::LoadInst>(v1)) &&
+         load->getPointerOperand() == vectorPtr)
+      v1SourceIsDest = true;
+    if ((load = llvm::dyn_cast<llvm::LoadInst>(v2)) &&
+         load->getPointerOperand() == vectorPtr)
+      v2SourceIsDest = true;
+
+    // Get mask indices that don't correspond to the destination vector
+    stack<unsigned> indices;
+    for (unsigned i = 0; i < maskSize; i++)
+    {
+      int idx = shuffle->getMaskValue(i);
+
+      // Skip undef indices
+      if (idx == -1)
+        continue;
+
+      // Check if source is the store destination
+      bool sourceIsDest =
+        ((unsigned)idx < v1->getType()->getVectorNumElements() ?
+          v1SourceIsDest : v2SourceIsDest);
+
+      // If destination is used in non-identity position, leave shuffle as is
+      if (sourceIsDest && (unsigned)idx != i)
+        return;
+
+      // Add non-destination index
+      if (!sourceIsDest)
+        indices.push(i);
+    }
+
+    // Check if destination is actually used as a source in the mask
+    if (indices.size() == maskSize)
+    {
+      // Check for any unused loads with the same address as the store
+      // These would usually be caught by DCE, but if optimisations are
+      // disabled we need to prune these manually
+      list<llvm::LoadInst*> lvalueloads;
+      for (auto user  = vectorPtr->user_begin();
+                user != vectorPtr->user_end() ;
+                user++)
+      {
+        if (auto load = llvm::dyn_cast<llvm::LoadInst>(*user))
+        {
+          if (load->getNumUses() == 0)
+            lvalueloads.push_back(load);
+        }
+      }
+      for (auto load = lvalueloads.begin(); load != lvalueloads.end(); load++)
+      {
+        (*load)->eraseFromParent();
+      }
+
+      return;
+    }
+
+    // Create a scalar store for each shuffle index
+    while (!indices.empty())
+    {
+      unsigned index = indices.top();
+      indices.pop();
+
+      // Create GEP for scalar value
+      llvm::GetElementPtrInst *scalarPtr = NULL;
+      if (auto gep = llvm::dyn_cast<llvm::GetElementPtrInst>(vectorPtr))
+      {
+        // Create GEP from existing GEP
+        std::vector<llvm::Value*> gepIndices;
+        for (auto idx = gep->idx_begin(); idx != gep->idx_end(); idx++)
+        {
+          gepIndices.push_back(*idx);
+        }
+        gepIndices.push_back(llvm::ConstantInt::getSigned(gepIndexType, index));
+        scalarPtr = llvm::GetElementPtrInst::Create(
+#if LLVM_VERSION > 36
+          gep->getPointerOperandType()->getPointerElementType(),
+#endif
+          gep->getPointerOperand(), gepIndices);
+      }
+      else
+      {
+        // Create GEP from non-GEP pointer
+        std::vector<llvm::Value*> gepIndices;
+        gepIndices.push_back(llvm::ConstantInt::getSigned(gepIndexType, 0));
+        gepIndices.push_back(llvm::ConstantInt::getSigned(gepIndexType, index));
+        scalarPtr = llvm::GetElementPtrInst::Create(
+#if LLVM_VERSION > 36
+          vectorPtr->getType()->getPointerElementType(),
+#endif
+          vectorPtr, gepIndices);
+      }
+      scalarPtr->setDebugLoc(store->getDebugLoc());
+      scalarPtr->insertAfter(store);
+
+      // Get source vector and index
+      unsigned idx   = shuffle->getMaskValue(index);
+      unsigned v1num = v1->getType()->getVectorNumElements();
+      llvm::Value *src = v1;
+      if (idx >= v1num)
+      {
+        idx -= v1num;
+        src = v2;
+      }
+
+      // Create direct scalar store
+      if (auto cnst = llvm::dyn_cast<llvm::ConstantVector>(src))
+      {
+        // If source is a constant, extract scalar constant
+        src = cnst->getAggregateElement(idx);
+
+        llvm::StoreInst *scalarStore = new llvm::StoreInst(
+          src, scalarPtr, store->isVolatile(),
+          getTypeAlignment(src->getType()));
+        scalarStore->setDebugLoc(store->getDebugLoc());
+        scalarStore->insertAfter(scalarPtr);
+      }
+      else
+      {
+        // If extracting from a shuffle, trace back to last non-shuffle
+        while (auto shfl = llvm::dyn_cast<llvm::ShuffleVectorInst>(src))
+        {
+          llvm::Value *v1 = shfl->getOperand(0);
+          llvm::Value *v2 = shfl->getOperand(1);
+          unsigned v1num  = v1->getType()->getVectorNumElements();
+
+          // Get source vector and index
+          idx = shfl->getMaskValue(idx);
+          src = v1;
+          if (idx >= v1num)
+          {
+            idx -= v1num;
+            src = v2;
+          }
+        }
+
+        llvm::ExtractElementInst *extract = llvm::ExtractElementInst::Create(
+          src, llvm::ConstantInt::getSigned(gepIndexType, idx));
+        extract->setDebugLoc(shuffle->getDebugLoc());
+        extract->insertAfter(scalarPtr);
+
+        llvm::StoreInst *scalarStore = new llvm::StoreInst(
+          extract, scalarPtr, store->isVolatile(),
+          getTypeAlignment(extract->getType()));
+        scalarStore->setDebugLoc(store->getDebugLoc());
+        scalarStore->insertAfter(extract);
+      }
+    }
+
+    // Prune old store and dead any code
+    store->eraseFromParent();
+    pruneDeadCode(shuffle);
+  }
+}
+
 void Program::stripDebugIntrinsics()
 {
   // Get list of llvm.dbg intrinsics
   set<llvm::Instruction*> intrinsics;
   for (llvm::Module::iterator F = m_module->begin(); F != m_module->end(); F++)
   {
-    for (llvm::inst_iterator I = inst_begin(F), E = inst_end(F); I != E; I++)
+    llvm::Function *f = &*F;
+    for (llvm::inst_iterator I = inst_begin(f), E = inst_end(f); I != E; I++)
     {
       if (I->getOpcode() == llvm::Instruction::Call)
       {
@@ -722,7 +1090,6 @@ void Program::stripDebugIntrinsics()
   set<llvm::Instruction*>::iterator itr;
   for (itr = intrinsics.begin(); itr != intrinsics.end(); itr++)
   {
-    (*itr)->removeFromParent();
-    delete *itr;
+    (*itr)->eraseFromParent();
   }
 }
