@@ -1,13 +1,16 @@
 // WorkItem.cpp (Oclgrind)
-// Copyright (c) 2013-2015, James Price and Simon McIntosh-Smith,
+// Copyright (c) 2013-2016, James Price and Simon McIntosh-Smith,
 // University of Bristol. All rights reserved.`
 //
 // This program is provided under a three-clause BSD license. For full
 // license terms please see the LICENSE file distributed with this
 // source code.
 
+#include "config.h"
 #include "common.h"
 
+#include <math.h>
+
 #include "llvm/IR/DebugInfo.h"
 #include "llvm/IR/GlobalVariable.h"
 #include "llvm/IR/InstrTypes.h"
@@ -30,9 +33,10 @@ using namespace std;
 
 struct WorkItem::Position
 {
-  llvm::Function::const_iterator       prevBlock;
-  llvm::Function::const_iterator       currBlock;
-  llvm::Function::const_iterator       nextBlock;
+  bool hasBegun;
+  const llvm::BasicBlock *             prevBlock;
+  const llvm::BasicBlock *             currBlock;
+  const llvm::BasicBlock *             nextBlock;
   llvm::BasicBlock::const_iterator     currInst;
   std::stack<const llvm::Instruction*> callStack;
   std::stack< std::list<size_t> >      allocations;
@@ -67,21 +71,48 @@ WorkItem::WorkItem(const KernelInvocation *kernelInvocation,
   // Set initial number of values to store based on cache
   m_values.resize(m_cache->getNumValues());
 
-  m_privateMemory = kernel->getPrivateMemory()->clone();
+  m_privateMemory = new Memory(AddrSpacePrivate, sizeof(size_t)==8 ? 32 : 16,
+                               m_context);
 
-  // Initialise kernel arguments
-  TypedValueMap::const_iterator argItr;
-  for (argItr = kernel->args_begin(); argItr != kernel->args_end(); argItr++)
+  // Initialise kernel arguments and global variables
+  for (auto value  = kernel->values_begin();
+            value != kernel->values_end();
+            value++)
   {
-    setValue(argItr->first, m_pool.clone(argItr->second));
+    pair<unsigned,unsigned> size = getValueSize(value->first);
+    TypedValue v = {
+      size.first,
+      size.second,
+      m_pool.alloc(size.first*size.second)
+    };
+
+    const llvm::Type *type = value->first->getType();
+    if (type->isPointerTy() &&
+        type->getPointerAddressSpace() == AddrSpacePrivate)
+    {
+      size_t sz = value->second.size*value->second.num;
+      v.setPointer(m_privateMemory->allocateBuffer(sz, 0, value->second.data));
+    }
+    else if (type->isPointerTy() &&
+             type->getPointerAddressSpace() == AddrSpaceLocal)
+    {
+      v.setPointer(m_workGroup->getLocalMemoryAddress(value->first));
+    }
+    else
+    {
+      memcpy(v.data, value->second.data, v.size*v.num);
+    }
+
+    setValue(value->first, v);
   }
 
   // Initialize interpreter state
   m_state    = READY;
   m_position = new Position;
+  m_position->hasBegun = false;
   m_position->prevBlock = NULL;
   m_position->nextBlock = NULL;
-  m_position->currBlock = kernel->getFunction()->begin();
+  m_position->currBlock = &*kernel->getFunction()->begin();
   m_position->currInst = m_position->currBlock->begin();
 }
 
@@ -298,19 +329,19 @@ void WorkItem::execute(const llvm::Instruction *instruction)
   m_context->notifyInstructionExecuted(this, instruction, result);
 }
 
-TypedValue WorkItem::getValue(const llvm::Value *key) const
-{
-  return m_values[m_cache->getValueID(key)];
-}
-
 const stack<const llvm::Instruction*>& WorkItem::getCallStack() const
 {
   return m_position->callStack;
 }
 
+const llvm::BasicBlock* WorkItem::getCurrentBlock() const
+{
+  return m_position->currBlock;
+}
+
 const llvm::Instruction* WorkItem::getCurrentInstruction() const
 {
-  return m_position->currInst;
+  return &*m_position->currInst;
 }
 
 Size3 WorkItem::getGlobalID() const
@@ -416,6 +447,11 @@ TypedValue WorkItem::getOperand(const llvm::Value *operand) const
   assert(false);
 }
 
+const llvm::BasicBlock* WorkItem::getPreviousBlock() const
+{
+  return m_position->prevBlock;
+}
+
 Memory* WorkItem::getPrivateMemory() const
 {
   return m_privateMemory;
@@ -426,6 +462,11 @@ WorkItem::State WorkItem::getState() const
   return m_state;
 }
 
+TypedValue WorkItem::getValue(const llvm::Value *key) const
+{
+  return m_values[m_cache->getValueID(key)];
+}
+
 const unsigned char* WorkItem::getValueData(const llvm::Value *value) const
 {
   if (!hasValue(value))
@@ -437,13 +478,27 @@ const unsigned char* WorkItem::getValueData(const llvm::Value *value) const
 
 const llvm::Value* WorkItem::getVariable(std::string name) const
 {
+  // Check private variables
   VariableMap::const_iterator itr;
   itr = m_variables.find(name);
-  if (itr == m_variables.end())
+  if (itr != m_variables.end())
+    return itr->second;
+
+  // Check global variables
+  string globalName = m_position->currBlock->getParent()->getName();
+  globalName += ".";
+  globalName += name;
+  const llvm::Module *module =
+    m_kernelInvocation->getKernel()->getFunction()->getParent();
+  for (auto global = module->global_begin();
+            global != module->global_end();
+            global++)
   {
-    return NULL;
+    if (global->getName() == globalName)
+      return &*global;
   }
-  return itr->second;
+
+  return NULL;
 }
 
 const WorkGroup* WorkItem::getWorkGroup() const
@@ -478,18 +533,18 @@ bool WorkItem::printVariable(string name) const
   }
 
   // Get variable value
-  TypedValue result = getValue(value);
+  TypedValue result = getOperand(value);
   const llvm::Type *type = value->getType();
 
-  if (((const llvm::Instruction*)value)->getOpcode()
-       == llvm::Instruction::Alloca)
+  if (value->getValueID() == llvm::Value::GlobalVariableVal ||
+      ((const llvm::Instruction*)value)->getOpcode()
+         == llvm::Instruction::Alloca)
   {
-    // If value is alloca result, look-up data at address
-    const llvm::Type *elemType = value->getType()->getPointerElementType();
+    // If value is alloca or global variable, look-up data at address
     size_t address = result.getPointer();
-
-    unsigned char *data = (unsigned char*)m_privateMemory->getPointer(address);
-    printTypedData(elemType, data);
+    Memory *memory = getMemory(value->getType()->getPointerAddressSpace());
+    unsigned char *data = (unsigned char*)memory->getPointer(address);
+    printTypedData(value->getType()->getPointerElementType(), data);
   }
   else
   {
@@ -508,8 +563,14 @@ WorkItem::State WorkItem::step()
 {
   assert(m_state == READY);
 
+  if (!m_position->hasBegun)
+  {
+    m_position->hasBegun = true;
+    m_context->notifyWorkItemBegin(this);
+  }
+
   // Execute the next instruction
-  execute(m_position->currInst);
+  execute(&*m_position->currInst);
 
   // Check if we've reached the end of the block
   if (++m_position->currInst == m_position->currBlock->end() ||
@@ -525,6 +586,9 @@ WorkItem::State WorkItem::step()
     }
   }
 
+  if (m_state == FINISHED)
+    m_context->notifyWorkItemComplete(this);
+
   return m_state;
 }
 
@@ -663,9 +727,9 @@ INSTRUCTION(call)
   // Check if function has definition
   if (!function->isDeclaration())
   {
-    m_position->callStack.push(m_position->currInst);
+    m_position->callStack.push(&*m_position->currInst);
     m_position->allocations.push(list<size_t>());
-    m_position->nextBlock = function->begin();
+    m_position->nextBlock = &*function->begin();
 
     // Set function arguments
     llvm::Function::const_arg_iterator argItr;
@@ -673,7 +737,30 @@ INSTRUCTION(call)
          argItr != function->arg_end(); argItr++)
     {
       const llvm::Value *arg = callInst->getArgOperand(argItr->getArgNo());
-      setValue(argItr, m_pool.clone(getOperand(arg)));
+      TypedValue value = getOperand(arg);
+
+      if (argItr->hasByValAttr())
+      {
+        // Make new copy of value in private memory
+        void *data = m_privateMemory->getPointer(value.getPointer());
+        size_t size = getTypeSize(argItr->getType()->getPointerElementType());
+        size_t ptr  = m_privateMemory->allocateBuffer(size, 0, (uint8_t*)data);
+        m_position->allocations.top().push_back(ptr);
+
+        // Pass new allocation to function
+        TypedValue address =
+        {
+          sizeof(size_t),
+          1,
+          m_pool.alloc(sizeof(size_t))
+        };
+        address.setPointer(ptr);
+        setValue(&*argItr, address);
+      }
+      else
+      {
+        setValue(&*argItr, m_pool.clone(value));
+      }
     }
 
     return;
@@ -780,20 +867,19 @@ INSTRUCTION(fcmp)
       r = a <= b;
       break;
     case llvm::CmpInst::FCMP_FALSE:
+    case llvm::CmpInst::FCMP_UNO:
       r = false;
       break;
     case llvm::CmpInst::FCMP_TRUE:
-      r = true;
-      break;
     case llvm::CmpInst::FCMP_ORD:
-    case llvm::CmpInst::FCMP_UNO:
+      r = true;
       break;
     default:
       FATAL_ERROR("Unsupported FCmp predicate: %d", pred);
     }
 
     // Deal with NaN operands
-    if (::isnan(a) || ::isnan(b))
+    if (std::isnan(a) || std::isnan(b))
     {
       r = !llvm::CmpInst::isOrdered(pred);
     }
@@ -1057,10 +1143,14 @@ INSTRUCTION(load)
 {
   const llvm::LoadInst *loadInst = (const llvm::LoadInst*)instruction;
   unsigned addressSpace = loadInst->getPointerAddressSpace();
-  size_t address = getOperand(loadInst->getPointerOperand()).getPointer();
+  const llvm::Value *opPtr = loadInst->getPointerOperand();
+  size_t address = getOperand(opPtr).getPointer();
 
   // Check address is correctly aligned
-  if (address & (loadInst->getAlignment()-1))
+  unsigned alignment = loadInst->getAlignment();
+  if (!alignment)
+    alignment = getTypeAlignment(opPtr->getType()->getPointerElementType());
+  if (address & (alignment-1))
   {
     m_context->logError("Invalid memory load - source pointer is "
                         "not aligned to the pointed type");
@@ -1116,7 +1206,8 @@ INSTRUCTION(ret)
 
   if (!m_position->callStack.empty())
   {
-    m_position->currInst = m_position->callStack.top();
+    m_position->currInst =
+      llvm::BasicBlock::const_iterator(m_position->callStack.top());
     m_position->currBlock = m_position->currInst->getParent();
     m_position->callStack.pop();
 
@@ -1124,7 +1215,7 @@ INSTRUCTION(ret)
     const llvm::Value *returnVal = retInst->getReturnValue();
     if (returnVal)
     {
-      setValue(m_position->currInst, m_pool.clone(getOperand(returnVal)));
+      setValue(&*m_position->currInst, m_pool.clone(getOperand(returnVal)));
     }
 
     // Clear stack allocations
@@ -1141,7 +1232,6 @@ INSTRUCTION(ret)
     m_position->nextBlock = NULL;
     m_state = FINISHED;
     m_workGroup->notifyFinished(this);
-    m_context->notifyWorkItemComplete(this);
   }
 }
 
@@ -1270,10 +1360,14 @@ INSTRUCTION(store)
 {
   const llvm::StoreInst *storeInst = (const llvm::StoreInst*)instruction;
   unsigned addressSpace = storeInst->getPointerAddressSpace();
-  size_t address = getOperand(storeInst->getPointerOperand()).getPointer();
+  const llvm::Value *opPtr = storeInst->getPointerOperand();
+  size_t address = getOperand(opPtr).getPointer();
 
   // Check address is correctly aligned
-  if (address & (storeInst->getAlignment()-1))
+  unsigned alignment = storeInst->getAlignment();
+  if (!alignment)
+    alignment = getTypeAlignment(opPtr->getType()->getPointerElementType());
+  if (address & (alignment-1))
   {
     m_context->logError("Invalid memory store - source pointer is "
                         "not aligned to the pointed type");
@@ -1322,7 +1416,11 @@ INSTRUCTION(uitofp)
   TypedValue op = getOperand(instruction->getOperand(0));
   for (unsigned i = 0; i < result.num; i++)
   {
-    result.setFloat(op.getUInt(i), i);
+    uint64_t in = op.getUInt(i);
+    if (result.size == 4)
+      result.setFloat(in ? (float)in : 0.f, i);
+    else
+      result.setFloat(in ? (double)in : 0.0, i);
   }
 }
 
@@ -1365,7 +1463,7 @@ InterpreterCache::InterpreterCache(llvm::Function *kernel)
   llvm::Module::const_global_iterator G;
   for (G = module->global_begin(); G != module->global_end(); G++)
   {
-    addValueID(G);
+    addValueID(&*G);
   }
 
 
@@ -1385,7 +1483,7 @@ InterpreterCache::InterpreterCache(llvm::Function *kernel)
     llvm::Function::arg_iterator A;
     for (A = function->arg_begin(); A != function->arg_end(); A++)
     {
-      addValueID(A);
+      addValueID(&*A);
     }
 
     // Iterate through instructions in function
@@ -1573,8 +1671,6 @@ bool InterpreterCache::hasValue(const llvm::Value *value) const
 
 void InterpreterCache::addOperand(const llvm::Value *operand)
 {
-  addValueID(operand);
-
   // Resolve constants
   if (operand->getValueID() == llvm::Value::UndefValueVal            ||
       operand->getValueID() == llvm::Value::ConstantAggregateZeroVal ||
@@ -1595,8 +1691,7 @@ void InterpreterCache::addOperand(const llvm::Value *operand)
     const llvm::ConstantExpr *expr = (const llvm::ConstantExpr*)operand;
     if (!m_constExpressions.count(expr))
     {
-      for (llvm::User::const_op_iterator O = expr->op_begin();
-           O != expr->op_end(); O++)
+      for (auto O = expr->op_begin(); O != expr->op_end(); O++)
       {
         addOperand(*O);
       }
@@ -1604,57 +1699,8 @@ void InterpreterCache::addOperand(const llvm::Value *operand)
       // TODO: Resolve actual value?
     }
   }
-}
-
-
-//////////////////////////
-// WorkItem::MemoryPool //
-//////////////////////////
-
-WorkItem::MemoryPool::MemoryPool(size_t blockSize) : m_blockSize(blockSize)
-{
-  // Force first allocation to create new block
-  m_offset = m_blockSize;
-}
-
-WorkItem::MemoryPool::~MemoryPool()
-{
-  list<unsigned char*>::iterator itr;
-  for (itr = m_blocks.begin(); itr != m_blocks.end(); itr++)
+  else
   {
-    delete[] *itr;
+    addValueID(operand);
   }
 }
-
-unsigned char* WorkItem::MemoryPool::alloc(size_t size)
-{
-  // Check if requested size larger than block size
-  if (size > m_blockSize)
-  {
-    // Oversized buffers allocated separately from main pool
-    unsigned char *buffer = new unsigned char[size];
-    m_blocks.push_back(buffer);
-    return buffer;
-  }
-
-  // Check if enough space in current block
-  if (m_offset + size > m_blockSize)
-  {
-    // Allocate new block
-    m_blocks.push_front(new unsigned char[m_blockSize]);
-    m_offset = 0;
-  }
-  unsigned char *buffer = m_blocks.front() + m_offset;
-  m_offset += size;
-  return buffer;
-}
-
-TypedValue WorkItem::MemoryPool::clone(const TypedValue& source)
-{
-  TypedValue dest;
-  dest.size = source.size;
-  dest.num = source.num;
-  dest.data = alloc(dest.size*dest.num);
-  memcpy(dest.data, source.data, dest.size*dest.num);
-  return dest;
-}
