@@ -1,5 +1,5 @@
 // clc.h (Oclgrind)
-// Copyright (c) 2013-2015, James Price and Simon McIntosh-Smith,
+// Copyright (c) 2013-2016, James Price and Simon McIntosh-Smith,
 // University of Bristol. All rights reserved.
 //
 // This program is provided under a three-clause BSD license. For full
@@ -421,37 +421,41 @@ BUILTIN_1ARG_FLOATS(fast_normalize);
 // Image Functions //
 /////////////////////
 
-size_t __OVERLOAD__ get_image_array_size(image1d_array_t image);
-size_t __OVERLOAD__ get_image_array_size(image2d_array_t image);
+#define IMAGE_QUERY(ret, name, type) \
+  ret __OVERLOAD__ name(read_only type image); \
+  ret __OVERLOAD__ name(write_only type image)
 
-int __OVERLOAD__ get_image_channel_data_type(image1d_t image);
-int __OVERLOAD__ get_image_channel_data_type(image1d_buffer_t image);
-int __OVERLOAD__ get_image_channel_data_type(image1d_array_t image);
-int __OVERLOAD__ get_image_channel_data_type(image2d_t image);
-int __OVERLOAD__ get_image_channel_data_type(image2d_array_t image);
-int __OVERLOAD__ get_image_channel_data_type(image3d_t image);
+IMAGE_QUERY(size_t, get_image_array_size, image1d_array_t);
+IMAGE_QUERY(size_t, get_image_array_size, image2d_array_t);
 
-int __OVERLOAD__ get_image_channel_order(image1d_t image);
-int __OVERLOAD__ get_image_channel_order(image1d_buffer_t image);
-int __OVERLOAD__ get_image_channel_order(image1d_array_t image);
-int __OVERLOAD__ get_image_channel_order(image2d_t image);
-int __OVERLOAD__ get_image_channel_order(image2d_array_t image);
-int __OVERLOAD__ get_image_channel_order(image3d_t image);
+IMAGE_QUERY(int, get_image_channel_data_type, image1d_t);
+IMAGE_QUERY(int, get_image_channel_data_type, image1d_buffer_t);
+IMAGE_QUERY(int, get_image_channel_data_type, image1d_array_t);
+IMAGE_QUERY(int, get_image_channel_data_type, image2d_t);
+IMAGE_QUERY(int, get_image_channel_data_type, image2d_array_t);
+IMAGE_QUERY(int, get_image_channel_data_type, image3d_t);
 
-int2 __OVERLOAD__ get_image_dim(image2d_t image);
-int2 __OVERLOAD__ get_image_dim(image2d_array_t image);
-int4 __OVERLOAD__ get_image_dim(image3d_t image);
+IMAGE_QUERY(int, get_image_channel_order, image1d_t);
+IMAGE_QUERY(int, get_image_channel_order, image1d_buffer_t);
+IMAGE_QUERY(int, get_image_channel_order, image1d_array_t);
+IMAGE_QUERY(int, get_image_channel_order, image2d_t);
+IMAGE_QUERY(int, get_image_channel_order, image2d_array_t);
+IMAGE_QUERY(int, get_image_channel_order, image3d_t);
 
-int __OVERLOAD__ get_image_depth(image3d_t image);
-int __OVERLOAD__ get_image_height(image2d_t image);
-int __OVERLOAD__ get_image_height(image2d_array_t image);
-int __OVERLOAD__ get_image_height(image3d_t image);
-int __OVERLOAD__ get_image_width(image1d_t image);
-int __OVERLOAD__ get_image_width(image1d_buffer_t image);
-int __OVERLOAD__ get_image_width(image1d_array_t image);
-int __OVERLOAD__ get_image_width(image2d_t image);
-int __OVERLOAD__ get_image_width(image2d_array_t image);
-int __OVERLOAD__ get_image_width(image3d_t image);
+IMAGE_QUERY(int2, get_image_dim, image2d_t);
+IMAGE_QUERY(int2, get_image_dim, image2d_array_t);
+IMAGE_QUERY(int4, get_image_dim, image3d_t);
+
+IMAGE_QUERY(int, get_image_depth, image3d_t);
+IMAGE_QUERY(int, get_image_height, image2d_t);
+IMAGE_QUERY(int, get_image_height, image2d_array_t);
+IMAGE_QUERY(int, get_image_height, image3d_t);
+IMAGE_QUERY(int, get_image_width, image1d_t);
+IMAGE_QUERY(int, get_image_width, image1d_buffer_t);
+IMAGE_QUERY(int, get_image_width, image1d_array_t);
+IMAGE_QUERY(int, get_image_width, image2d_t);
+IMAGE_QUERY(int, get_image_width, image2d_array_t);
+IMAGE_QUERY(int, get_image_width, image3d_t);
 
 float4 __OVERLOAD__ read_imagef(image1d_t, int);
 float4 __OVERLOAD__ read_imagef(image1d_buffer_t, int);
@@ -507,21 +511,21 @@ uint4 __OVERLOAD__ read_imageui(image2d_array_t, sampler_t, float4);
 uint4 __OVERLOAD__ read_imageui(image3d_t, sampler_t, int4);
 uint4 __OVERLOAD__ read_imageui(image3d_t, sampler_t, float4);
 
-void __OVERLOAD__ write_imagef(image1d_t, int, float4);
-void __OVERLOAD__ write_imagef(image1d_array_t, int2, float4);
-void __OVERLOAD__ write_imagef(image2d_t, int2, float4);
-void __OVERLOAD__ write_imagef(image2d_array_t, int4, float4);
-void __OVERLOAD__ write_imagef(image3d_t, int4, float4);
-void __OVERLOAD__ write_imagei(image1d_t, int, int4);
-void __OVERLOAD__ write_imagei(image1d_array_t, int2, int4);
-void __OVERLOAD__ write_imagei(image2d_t, int2, int4);
-void __OVERLOAD__ write_imagei(image2d_array_t, int4, int4);
-void __OVERLOAD__ write_imagei(image3d_t, int4, int4);
-void __OVERLOAD__ write_imageui(image1d_t, int, uint4);
-void __OVERLOAD__ write_imageui(image1d_array_t, int2, uint4);
-void __OVERLOAD__ write_imageui(image2d_t, int2, uint4);
-void __OVERLOAD__ write_imageui(image2d_array_t, int4, uint4);
-void __OVERLOAD__ write_imageui(image3d_t, int4, uint4);
+void __OVERLOAD__ write_imagef(write_only image1d_t, int, float4);
+void __OVERLOAD__ write_imagef(write_only image1d_array_t, int2, float4);
+void __OVERLOAD__ write_imagef(write_only image2d_t, int2, float4);
+void __OVERLOAD__ write_imagef(write_only image2d_array_t, int4, float4);
+void __OVERLOAD__ write_imagef(write_only image3d_t, int4, float4);
+void __OVERLOAD__ write_imagei(write_only image1d_t, int, int4);
+void __OVERLOAD__ write_imagei(write_only image1d_array_t, int2, int4);
+void __OVERLOAD__ write_imagei(write_only image2d_t, int2, int4);
+void __OVERLOAD__ write_imagei(write_only image2d_array_t, int4, int4);
+void __OVERLOAD__ write_imagei(write_only image3d_t, int4, int4);
+void __OVERLOAD__ write_imageui(write_only image1d_t, int, uint4);
+void __OVERLOAD__ write_imageui(write_only image1d_array_t, int2, uint4);
+void __OVERLOAD__ write_imageui(write_only image2d_t, int2, uint4);
+void __OVERLOAD__ write_imageui(write_only image2d_array_t, int4, uint4);
+void __OVERLOAD__ write_imageui(write_only image3d_t, int4, uint4);
 
 
 ///////////////////////
@@ -890,7 +894,9 @@ VLOADSTORE(double);
   VSTORE_HALF_ADDRSPACE(func##_rtn, type);
 #define VSTORE_HALF_WIDTH(n)                    \
   VSTORE_HALF_ROUND(vstore_half##n, float##n);  \
-  VSTORE_HALF_ROUND(vstorea_half##n, float##n);
+  VSTORE_HALF_ROUND(vstorea_half##n, float##n); \
+  VSTORE_HALF_ROUND(vstore_half##n, double##n); \
+  VSTORE_HALF_ROUND(vstorea_half##n, double##n);
 #define VLOADSTORE_HALF_WIDTH(n) \
   VLOAD_HALF_WIDTH(n);           \
   VSTORE_HALF_WIDTH(n);
