@@ -1,11 +1,12 @@
 // common.cpp (Oclgrind)
-// Copyright (c) 2013-2015, James Price and Simon McIntosh-Smith,
+// Copyright (c) 2013-2016, James Price and Simon McIntosh-Smith,
 // University of Bristol. All rights reserved.
 //
 // This program is provided under a three-clause BSD license. For full
 // license terms please see the LICENSE file distributed with this
 // source code.
 
+#include "config.h"
 #include "common.h"
 
 #if defined(_WIN32) && !defined(__MINGW32__)
@@ -24,19 +25,19 @@ using namespace std;
 
 namespace oclgrind
 {
-  _Size3_::_Size3_()
+  Size3::Size3()
   {
     x = y = z = 0;
   }
 
-  _Size3_::_Size3_(size_t _x, size_t _y, size_t _z)
+  Size3::Size3(size_t _x, size_t _y, size_t _z)
   {
     x = _x;
     y = _y;
     z = _z;
   }
 
-  _Size3_::_Size3_(size_t linear, _Size3_ dimensions)
+  Size3::Size3(size_t linear, Size3 dimensions)
   {
     x = linear % dimensions.x;
     y = (linear / dimensions.x) % dimensions.y;
@@ -55,6 +56,7 @@ namespace oclgrind
       return z;
     default:
       assert(false && "Size3 index out of range");
+      abort();
     }
   }
 
@@ -70,6 +72,7 @@ namespace oclgrind
       return z;
     default:
       assert(false && "Size3 index out of range");
+      abort();
     }
   }
 
@@ -78,6 +81,11 @@ namespace oclgrind
     return x == rhs.x && y == rhs.y && z == rhs.z;
   }
 
+  bool Size3::operator!=(const Size3& rhs) const
+  {
+    return x != rhs.x || y != rhs.y || z != rhs.z;
+  }
+
   ostream& operator<<(ostream& stream, const Size3& size)
   {
     stream << dec    << "("
@@ -209,13 +217,68 @@ namespace oclgrind
     }
   }
 
+  ostream& operator<<(ostream& stream, const TypedValue& tv)
+  {
+    if(tv.data)
+    {
+      if(tv.num > 1)
+      {
+        stream << "(";
+      }
+
+      for(unsigned n = 0; n < tv.num; ++n)
+      {
+        for(int i = tv.size - 1; i >= 0; --i)
+        {
+          stream << hex << uppercase << setw(2) << setfill('0')
+                 << (int)*(tv.data + tv.size * n + i);
+        }
+
+        if(n != tv.num - 1)
+        {
+          stream << ",";
+        }
+      }
+
+      if(tv.num > 1)
+      {
+        stream << ")";
+      }
+    }
+    else
+    {
+      stream << "NULL";
+    }
+
+    return stream;
+  }
+
+  bool TypedValue::operator==(const TypedValue& rhs) const
+  {
+    return (size == rhs.size) && (num == rhs.num) &&
+           (memcmp(data, rhs.data, size*num) == 0);
+  }
+
+  bool TypedValue::operator!=(const TypedValue& rhs) const
+  {
+    return (size != rhs.size) || (num != rhs.num) ||
+           (memcmp(data, rhs.data, size*num) != 0);
+  }
+
   TypedValue TypedValue::clone() const
   {
     TypedValue result;
     result.size = size;
     result.num  = num;
-    result.data = new unsigned char[size*num];
-    memcpy(result.data, data, size*num);
+    if (data)
+    {
+      result.data = new unsigned char[size*num];
+      memcpy(result.data, data, size*num);
+    }
+    else
+    {
+      result.data = NULL;
+    }
     return result;
   }
 
@@ -260,10 +323,27 @@ namespace oclgrind
     switch (type->getTypeID())
     {
     case llvm::Type::IntegerTyID:
-      memcpy(data,
-             ((llvm::ConstantInt*)constant)->getValue().getRawData(),
-             size);
+    {
+      uint64_t ui = ((llvm::ConstantInt*)constant)->getZExtValue();
+      switch (size)
+      {
+      case 1:
+        *((uint8_t*)data) = ui;
+        break;
+      case 2:
+        *((uint16_t*)data) = ui;
+        break;
+      case 4:
+        *((uint32_t*)data) = ui;
+        break;
+      case 8:
+        *((uint64_t*)data) = ui;
+        break;
+      default:
+        FATAL_ERROR("Unsupported constant int size: %u bytes", size);
+      }
       break;
+    }
     case llvm::Type::FloatTyID:
     {
       *(float*)data =
@@ -328,13 +408,8 @@ namespace oclgrind
     const llvm::ConstantExpr *expr)
   {
     // Get operands
-    unsigned numOperands = expr->getNumOperands();
-    llvm::Value **valueOperands = new llvm::Value*[numOperands];
-    for (unsigned i = 0; i < numOperands; i++)
-    {
-      valueOperands[i] = expr->getOperand(i);
-    }
-    llvm::ArrayRef<llvm::Value*> operands(valueOperands, numOperands);
+    vector<llvm::Value*> valueOperands(expr->op_begin(), expr->op_end());
+    llvm::ArrayRef<llvm::Value*> operands(valueOperands);
 
     // Create instruction
     unsigned opcode = expr->getOpcode();
@@ -378,7 +453,7 @@ namespace oclgrind
       else
       {
 #if LLVM_VERSION > 36
-        return llvm::GetElementPtrInst::Create(expr->getType(),
+        return llvm::GetElementPtrInst::Create(nullptr,
                                                operands[0], operands.slice(1));
 #else
         return llvm::GetElementPtrInst::Create(operands[0], operands.slice(1));
@@ -387,9 +462,10 @@ namespace oclgrind
       }
     case llvm::Instruction::ICmp:
     case llvm::Instruction::FCmp:
-      return llvm::CmpInst::Create((llvm::Instruction::OtherOps)opcode,
-                                   expr->getPredicate(),
-                                   operands[0], operands[1]);
+      return llvm::CmpInst::Create(
+        (llvm::Instruction::OtherOps)opcode,
+        (llvm::CmpInst::Predicate)expr->getPredicate(),
+        operands[0], operands[1]);
     default:
       assert(expr->getNumOperands() == 2 && "Must be binary operator?");
 
@@ -425,10 +501,9 @@ namespace oclgrind
     }
   }
 
-  const llvm::ConstantInt* getMDOpAsConstInt(const llvm::MDOperand& op)
+  const llvm::ConstantInt* getMDAsConstInt(const llvm::Metadata *md)
   {
-    llvm::Metadata *md = op.get();
-    llvm::ConstantAsMetadata *cam =
+    const llvm::ConstantAsMetadata *cam =
       llvm::dyn_cast<llvm::ConstantAsMetadata>(md);
     if (!cam)
       return NULL;
@@ -461,7 +536,7 @@ namespace oclgrind
     }
 
     // Unreachable
-    assert(false);
+    abort();
   }
 
   unsigned getTypeSize(const llvm::Type *type)
@@ -517,16 +592,9 @@ namespace oclgrind
     }
     else
     {
-      // For some reason, getScalarSizeInBits is not const
-      llvm::Type* nonConstTy = const_cast<llvm::Type*>(type);
-
       // Round up for types that have a bit size not multiple of 8
       // like "bool".
-      unsigned ret = nonConstTy->getScalarSizeInBits() / 8;
-      if (nonConstTy->getScalarSizeInBits() % 8)
-        ret++;
-
-      return ret;
+      return (type->getScalarSizeInBits() + 7) >> 3;
     }
   }
 
@@ -584,7 +652,7 @@ namespace oclgrind
       numElements = 1;
     }
 
-    unsigned elemSize = bits >> 3;
+    unsigned elemSize = (bits+7) >> 3;
 
     // Special case for pointer types
     if (type->isPointerTy())
@@ -675,6 +743,20 @@ namespace oclgrind
     case llvm::Type::PointerTyID:
       cout << "0x" << hex << *(size_t*)data;
       break;
+    case llvm::Type::ArrayTyID:
+    {
+      const llvm::Type *elemType = type->getArrayElementType();
+      unsigned elemSize = getTypeSize(elemType);
+      cout << "{";
+      for (unsigned i = 0; i < type->getArrayNumElements(); i++)
+      {
+        if (i > 0)
+          cout << ",";
+        printTypedData(elemType, data+i*elemSize);
+      }
+      cout << "}";
+      break;
+    }
     default:
       cout << "(raw) 0x" << hex << uppercase << setfill('0');
       for (unsigned i = 0; i < size; i++)
@@ -709,4 +791,51 @@ namespace oclgrind
   {
     return runtime_error::what();
   }
+
+  MemoryPool::MemoryPool(size_t blockSize) : m_blockSize(blockSize)
+  {
+    // Force first allocation to create new block
+    m_offset = m_blockSize;
+  }
+
+  MemoryPool::~MemoryPool()
+  {
+    for (auto itr = m_blocks.begin(); itr != m_blocks.end(); itr++)
+    {
+      delete[] *itr;
+    }
+  }
+
+  uint8_t* MemoryPool::alloc(size_t size)
+  {
+    // Check if requested size larger than block size
+    if (size > m_blockSize)
+    {
+      // Oversized buffers allocated separately from main pool
+      unsigned char *buffer = new unsigned char[size];
+      m_blocks.push_back(buffer);
+      return buffer;
+    }
+
+    // Check if enough space in current block
+    if (m_offset + size > m_blockSize)
+    {
+      // Allocate new block
+      m_blocks.push_front(new unsigned char[m_blockSize]);
+      m_offset = 0;
+    }
+    uint8_t *buffer = m_blocks.front() + m_offset;
+    m_offset += size;
+    return buffer;
+  }
+
+  TypedValue MemoryPool::clone(const TypedValue& source)
+  {
+    TypedValue dest;
+    dest.size = source.size;
+    dest.num = source.num;
+    dest.data = alloc(dest.size*dest.num);
+    memcpy(dest.data, source.data, dest.size*dest.num);
+    return dest;
+  }
 }
