@@ -1,12 +1,11 @@
 // Simulation.cpp (Oclgrind)
-// Copyright (c) 2013-2015, James Price and Simon McIntosh-Smith,
+// Copyright (c) 2013-2016, James Price and Simon McIntosh-Smith,
 // University of Bristol. All rights reserved.
 //
 // This program is provided under a three-clause BSD license. For full
 // license terms please see the LICENSE file distributed with this
 // source code.
 
-#include "config.h"
 #include <cassert>
 #include <cmath>
 #include <iostream>
@@ -59,10 +58,13 @@ void Simulation::dumpArgument(DumpArg& arg)
   for (size_t i = 0; i < num; i++)
   {
     cout << "  " << arg.name << "[" << i << "] = ";
+    if (arg.hex)
+      cout << "0x" << setfill('0') << setw(sizeof(T)*2) << hex;
     if (sizeof(T) == 1)
       cout << (int)data[i];
     else
       cout << data[i];
+    cout << dec;
     cout << endl;
   }
   cout << endl;
@@ -267,6 +269,8 @@ void Simulation::parseArgument(size_t index)
   size_t typeSize = 0;
   bool null = false;
   bool dump = false;
+  bool hex = false;
+  bool noinit = false;
   string fill = "";
   string range = "";
   string name = m_kernel->getArgumentName(index).str();
@@ -357,9 +361,19 @@ void Simulation::parseArgument(size_t index)
     }
     else if (token == "hex")
     {
+      hex = true;
       m_lineBuffer.setf(ios_base::hex);
       m_lineBuffer.unsetf(ios_base::dec | ios_base::oct);
     }
+    else if (token == "noinit")
+    {
+      if (addrSpace != CL_KERNEL_ARG_ADDRESS_GLOBAL &&
+          addrSpace != CL_KERNEL_ARG_ADDRESS_CONSTANT)
+      {
+        throw "'noinit' only valid for buffer arguments";
+      }
+      noinit = true;
+    }
     else if (token == "null")
     {
       if (addrSpace != CL_KERNEL_ARG_ADDRESS_GLOBAL &&
@@ -429,7 +443,7 @@ void Simulation::parseArgument(size_t index)
   // Ensure size given
   if (null)
   {
-    if (size != -1 || !fill.empty() || !range.empty())
+    if (size != -1 || !fill.empty() || !range.empty() || noinit || dump)
     {
       throw "'null' not valid with other argument descriptors";
     }
@@ -482,10 +496,16 @@ void Simulation::parseArgument(size_t index)
     {
       throw "'dump' only valid for memory objects";
     }
-    if (null)
-    {
-      throw "'dump' not valid with 'null' specifier";
-    }
+  }
+
+  // Ensure only one initializer given
+  unsigned numInitializers = 0;
+  if (noinit) numInitializers++;
+  if (!fill.empty()) numInitializers++;
+  if (!range.empty()) numInitializers++;
+  if (numInitializers > 1)
+  {
+    throw "Multiple initializers present";
   }
 
   // Generate argument data
@@ -506,7 +526,8 @@ void Simulation::parseArgument(size_t index)
   {
     // Parse argument data
     unsigned char *data = new unsigned char[size];
-    if (!fill.empty())
+    if (noinit){}
+    else if (!fill.empty())
     {
       istringstream fillStream(fill);
       fillStream.copyfmt(m_lineBuffer);
@@ -591,7 +612,10 @@ void Simulation::parseArgument(size_t index)
       // Allocate buffer and store content
       Memory *globalMemory = m_context->getGlobalMemory();
       size_t address = globalMemory->allocateBuffer(size, flags);
-      globalMemory->store((unsigned char*)&data[0], address, size);
+      if (!address)
+        throw "Failed to allocate global memory";
+      if (!noinit)
+        globalMemory->store((unsigned char*)&data[0], address, size);
       value.data = new unsigned char[value.size];
       value.setPointer(address);
       delete[] data;
@@ -604,6 +628,7 @@ void Simulation::parseArgument(size_t index)
           size,
           type,
           name,
+          hex
         };
         m_dumpArguments.push_back(dump);
       }
