@@ -1,11 +1,12 @@
 // InteractiveDebugger.cpp (Oclgrind)
-// Copyright (c) 2013-2015, James Price and Simon McIntosh-Smith,
+// Copyright (c) 2013-2016, James Price and Simon McIntosh-Smith,
 // University of Bristol. All rights reserved.
 //
 // This program is provided under a three-clause BSD license. For full
 // license terms please see the LICENSE file distributed with this
 // source code.
 
+#include "config.h"
 #include "core/common.h"
 
 #include <iterator>
@@ -308,7 +309,7 @@ void InteractiveDebugger::printFunction(
       cout << ", ";
     }
     cout << argItr->getName().str() << "=";
-    m_kernelInvocation->getCurrentWorkItem()->printValue(argItr);
+    m_kernelInvocation->getCurrentWorkItem()->printValue(&*argItr);
   }
 
   cout << ") at line " << dec << getLineNumber(instruction) << endl;
@@ -867,19 +868,27 @@ bool InteractiveDebugger::print(vector<string> args)
         cout << "not found" << endl;
         return false;
       }
+
       const llvm::Type *ptrType = ptr->getType();
+      unsigned addrSpace = ptrType->getPointerAddressSpace();
 
       // Check for alloca instruction, in which case look at allocated type
       bool alloca = false;
+      if (ptr->getValueID() == llvm::Value::GlobalVariableVal)
+      {
+        ptrType = ptrType->getPointerElementType();
+      }
       if (ptr->getValueID() >= llvm::Value::InstructionVal &&
           ((llvm::Instruction*)ptr)->getOpcode() == llvm::Instruction::Alloca)
       {
         ptrType = ((const llvm::AllocaInst*)ptr)->getAllocatedType();
+        if (ptrType->isPointerTy())
+          addrSpace = ptrType->getPointerAddressSpace();
         alloca = true;
       }
 
       // Ensure type is a pointer
-      if (!ptrType->isPointerTy())
+      if (!ptrType->isPointerTy() && !ptrType->isArrayTy())
       {
         cout << "not a pointer" << endl;
         return false;
@@ -891,12 +900,12 @@ bool InteractiveDebugger::print(vector<string> args)
       {
         // Load base address from private memory
         workItem->getPrivateMemory()->load((unsigned char*)&base,
-                                                    base, sizeof(size_t));
+                                           base, sizeof(size_t));
       }
 
       // Get target memory object
       Memory *memory = NULL;
-      switch (ptrType->getPointerAddressSpace())
+      switch (addrSpace)
       {
       case AddrSpacePrivate:
         memory = workItem->getPrivateMemory();
@@ -932,7 +941,12 @@ bool InteractiveDebugger::print(vector<string> args)
     }
     else
     {
-      if (!workItem->printVariable(args[i]))
+      try
+      {
+        if (!workItem->printVariable(args[i]))
+          cout << "not found";
+      }
+      catch (FatalError err)
       {
         cout << "not found";
       }
